package com.gk_software.core.maven.mojo;

import com.gk_software.core.maven.utils.SrcFiles;
import org.apache.maven.execution.MavenSession;
import org.apache.maven.plugin.AbstractMojo;
import org.apache.maven.plugin.BuildPluginManager;
import org.apache.maven.plugin.MojoExecutionException;
import org.apache.maven.plugins.annotations.Component;
import org.apache.maven.plugins.annotations.Parameter;
import org.apache.maven.project.MavenProject;
import org.codehaus.plexus.util.FileUtils;
import com.gk_software.core.maven.utils.ClassData;

import static org.twdata.maven.mojoexecutor.MojoExecutor.*;

import com.gk_software.core.maven.utils.ModificationHelper;

import java.io.File;
import java.io.IOException;
import java.util.*;


public abstract class AbstractExecution extends AbstractMojo {
    @Parameter(defaultValue = "${project}", required = true, readonly = true)
    MavenProject project;

    @Parameter(required = true)
    private String oldVersionPath;

    @Parameter(required = true)
    private String newVersionPath;

    @Parameter(defaultValue = "")
    private String output;

    @Parameter(defaultValue = "${session}")
    private MavenSession session;

    @Component
    private BuildPluginManager buildPluginManager;

    @Parameter(property = "artifactItems")
    private List<ArtifactItem> artifactItems;


    protected static final String javaSuffix = "java";
    protected static final String resourceSuffix = "resources";
    protected static final String doSuffix = "do-model";
    private String[] projectSources = {javaSuffix, resourceSuffix, doSuffix};


    protected boolean checkProductData() {
        oldVersionPath = oldVersionPath.replace("/", "\\");
        newVersionPath = newVersionPath.replace("/", "\\");
        return checkProductDataOnPath(oldVersionPath, true) && checkProductDataOnPath(newVersionPath, false);
    }

    private boolean checkProductDataOnPath(String path, boolean isOldVersion) {
        File product = new File(path);
        if (product.exists()) {
            if (artifactItems.size() > 0) {
                if (isOldVersion)
                    getLog().warn("Artifact defined in configuration are ignored because old product directory already exists.");
                else
                    getLog().warn("Artifact defined in configuration are ignored because new product directory already exists.");
            }
            return true;
        }

        if (artifactItems.size() < 1) {
            getLog().error("Product was not found on path: " + path);
            getLog().warn("There are no defined artifacts");
            return false;
        }

        if (isOldVersion)
            getLog().info("Downloading old version of product");
        else
            getLog().info("Downloading new version of product");

        getLog().info("---------------------------------------");

            for (ArtifactItem item : artifactItems) {
                try {
                    if (isOldVersion)
                        downloadArtifact(item.getGroupId(), item.getArtifactId(), item.getClassifier(), item.getOldVersion(), path);
                    else
                        downloadArtifact(item.getGroupId(), item.getArtifactId(), item.getClassifier(), item.getNewVersion(), path);
                } catch (MojoExecutionException e) {
                    e.printStackTrace();
                    return false;
                }
            }

        System.out.println("\n");
        return true;
    }

    protected boolean checkProjectData() {
        LinkedList<String> srcPaths = new LinkedList<>();
        for (String path : projectSources) {
            String absPath = this.project.getBuild().getSourceDirectory().replace("java", path);
            File tmpSrcFile = new File(absPath);
            if (!tmpSrcFile.exists()) {
                getLog().warn(path + " is missing");
            } else {
                srcPaths.add(absPath);
            }
        }
        return srcPaths.size() != 0;

    }

    private LinkedList<File> listFilesFromFolder(final File parent) {
        Stack<File> stack = new Stack<>();
        stack.push(parent);
        LinkedList<File> files = new LinkedList<>();
        while (!stack.isEmpty()) {
            File child = stack.pop();
            if (child.isDirectory()) {
                for (File f : Objects.requireNonNull(child.listFiles())) stack.push(f);
            } else {
                files.add(child);
            }
        }
        return files;
    }

    private LinkedList<File> listFilesFromFolder(final File parent, String includes, String excludes) {
        String[] regExcludes = ModificationHelper.excludesIntoPath(excludes);
        String[] includesSuffixes = ModificationHelper.strToRegex(includes);
        Stack<File> stack = new Stack<>();
        stack.push(parent);
        LinkedList<File> files = new LinkedList<>();
        while (!stack.isEmpty()) {
            File child = stack.pop();
            if (child.isDirectory()) {
                for (File f : Objects.requireNonNull(child.listFiles())) stack.push(f);
            } else {
                for (String suffix : includesSuffixes) {
                    if (child.getName().endsWith(suffix)) {
                        if (!ModificationHelper.isAmongExcludes(regExcludes, child.getAbsolutePath())) {
                            files.add(child);
                        }
                        break;
                    }
                }

            }
        }
        return files;
    }

    private LinkedList<File> listFilesFromFolder(final File parent, String includes) {
        String[] includesSuffixes = ModificationHelper.strToRegex(includes);
        Stack<File> stack = new Stack<>();
        stack.push(parent);
        LinkedList<File> files = new LinkedList<>();
        while (!stack.isEmpty()) {
            File child = stack.pop();
            if (child.isDirectory()) {
                for (File f : Objects.requireNonNull(child.listFiles())) stack.push(f);
            } else {
                for (String suffix : includesSuffixes) {
                    if (child.getName().endsWith(suffix)) {
                        files.add(child);
                        break;
                    }
                }
            }
        }
        return files;
    }

    private LinkedList<File> listFilesFromFolderExcludes(final File parent, String excludes) {
        String[] regExcludes = ModificationHelper.excludesIntoPath(excludes);
        Stack<File> stack = new Stack<>();
        stack.push(parent);
        LinkedList<File> files = new LinkedList<>();
        while (!stack.isEmpty()) {
            File child = stack.pop();
            if (child.isDirectory()) {
                for (File f : Objects.requireNonNull(child.listFiles())) stack.push(f);
            } else {
                if (!ModificationHelper.isAmongExcludes(regExcludes, child.getAbsolutePath())) {
                    files.add(child);
                }
            }
        }
        return files;
    }

    private LinkedList<File> getFilesFromFolder(File folder, String includes, String excludes) {
        if (((includes == null) || (includes.compareTo("*") == 0)) && (excludes == null)) {
            return listFilesFromFolder(folder);
        }
        if ((includes != null) && (excludes != null)) {
            return listFilesFromFolder(folder, includes, excludes);
        } else if (includes != null) {
            return listFilesFromFolder(folder, includes);
        } else {
            return listFilesFromFolderExcludes(folder, excludes);
        }
    }

    protected void downloadArtifacts(String groupId, String artifactId, String classifier, String oldVersion, String newVersion) {
        if (classifier == null) classifier = "sources";
        try {
            downloadArtifact(groupId, artifactId, classifier, oldVersion, "${project.build.directory}/old");
            downloadArtifact(groupId, artifactId, classifier, newVersion, "${project.build.directory}/new");
        } catch (MojoExecutionException e) {
            e.printStackTrace();
        }

    }

    private void downloadArtifact(String groupId, String artifactId, String classifier, String version, String location) throws MojoExecutionException {
        executeMojo(
                plugin(
                        groupId("org.apache.maven.plugins"),
                        artifactId("maven-dependency-plugin")
                ),
                goal("unpack"),
                configuration(
                        element("artifactItems",
                                element("artifactItem",
                                        element("groupId", groupId),
                                        element("artifactId", artifactId),
                                        element("version", version),
                                        element("classifier", classifier),
                                        element("overWrite", "false"),
                                        element("outputDirectory", location)))
                ),
                executionEnvironment(
                        project,
                        session,
                        buildPluginManager
                )
        );
    }


    protected SrcFiles getDoModelFiles() {
        String resourceFolderPath = this.project.getBuild().getSourceDirectory();
        resourceFolderPath = resourceFolderPath.replace(javaSuffix, doSuffix);
        return new SrcFiles(resourceFolderPath, listFilesFromFolder(new File(resourceFolderPath)));
    }

    protected LinkedList<File> getOldVersionFiles() {
        File oldProduct = new File(this.oldVersionPath);
        return listFilesFromFolder(oldProduct);
    }

    protected LinkedList<File> getOldVersionFiles(String includes, String excludes) {
        File oldProduct = new File(this.oldVersionPath);
        return getFilesFromFolder(oldProduct, includes, excludes);
    }

    protected LinkedList<File> getNewVersionFiles(String includes, String excludes) {
        File newProduct = new File(this.newVersionPath);
        return getFilesFromFolder(newProduct, includes, excludes);
    }

    protected LinkedList<File> getNewVersionFiles() {
        File newProduct = new File(this.newVersionPath);
        return listFilesFromFolder(newProduct);
    }

    protected SrcFiles getProjectSrcFiles() {
        String path = this.project.getBuild().getSourceDirectory();
        return new SrcFiles(path, listFilesFromFolder(new File(path)));
    }

    protected LinkedList<File> getChangedProductFiles() {
        File oldFile = new File(this.oldVersionPath);
        File newFile = new File(this.newVersionPath);
        return findChangedProductFiles(oldFile, newFile);
    }

    protected SrcFiles getProjectResourceFiles() {
        String resourceFolderPath = this.project.getBuild().getSourceDirectory();
        resourceFolderPath = resourceFolderPath.replace(javaSuffix, resourceSuffix);
        return new SrcFiles(resourceFolderPath, listFilesFromFolder(new File(resourceFolderPath)));
    }


    protected Hashtable<String, LinkedList<File>> getAllSrcFiles() {
        Hashtable<String, LinkedList<File>> files = new Hashtable<>();
        for (String path : projectSources) {
            String absPath = this.project.getBuild().getSourceDirectory().replace("java", path);
            File tmpSrcFile = new File(absPath);
            if (!tmpSrcFile.exists()) {
            } else {
                files.put(absPath, listFilesFromFolder(tmpSrcFile));
            }
        }
        return files;
    }

    protected boolean checkIfChanged(Hashtable<String, ClassData> hashtable, String prjFileRelativePath, String path, LinkedList<File> changedPromoFiles, String[] excludes) {
        File newVFile = new File(this.getNewVersionPath() + path);
        if (!newVFile.exists()) {
            if (!ModificationHelper.isAmongExcludes(excludes, path)) {
                ModificationHelper.addToHashTable(hashtable, path, prjFileRelativePath, "does not exist in NEW version");
                return true;
            }
            return false;
        }
        for (File f : changedPromoFiles) {
            if (f.getAbsolutePath().endsWith(path)) {
                try {
                    if (!FileUtils.contentEquals(f, newVFile)) {
                        if (!ModificationHelper.isAmongExcludes(excludes, path)) {
                            ModificationHelper.addToHashTable(hashtable, path, prjFileRelativePath, "changed in NEW version");
                            return true;
                        }
                        return false;
                    }
                } catch (IOException e) {
                    System.out.println(e.getMessage());
                }
            }
        }
        return false;
    }

    private LinkedList<File> findChangedProductFiles(File oldV, File newV) {
        LinkedList<File> changedBetweenVersions = new LinkedList<>();
        Hashtable<String, File> oldFiles = this.getFilesFromFolder(oldV, oldVersionPath);
        Hashtable<String, File> newFiles = this.getFilesFromFolder(newV, newVersionPath);
        for (String relativePath : oldFiles.keySet()) {
            if (newFiles.containsKey(relativePath)) {
                try {
                    if (!FileUtils.contentEquals(oldFiles.get(relativePath), newFiles.get(relativePath))) {
                        changedBetweenVersions.add(oldFiles.get(relativePath));
                    }
                } catch (IOException e) {
                    System.out.println(e.getMessage());
                }
            }
        }
        return changedBetweenVersions;
    }

    private Hashtable<String, File> getFilesFromFolder(final File parent, String prefix) {
        Stack<File> stack = new Stack<>();
        stack.push(parent);
        Hashtable<String, File> files = new Hashtable<>();
        while (!stack.isEmpty()) {
            File child = stack.pop();
            if (child.isDirectory()) {
                for (File f : Objects.requireNonNull(child.listFiles())) stack.push(f);
            } else {
                String relativePath = ModificationHelper.getPathFromString(child.getAbsolutePath(), prefix);
                files.put(relativePath, child);
            }
        }
        return files;
    }


    public String getProjectResourcePath() {
        return this.project.getBuild().getSourceDirectory().replace(javaSuffix, resourceSuffix);
    }

    public String getOldVersionPath() {
        return this.oldVersionPath;
    }

    public String getNewVersionPath() {
        return this.newVersionPath;
    }

    public String getOutput() {
        return checkOutput();
    }

    private String checkOutput() {
        String tmpOutput = this.output;
        if (tmpOutput != null) {
            if (!tmpOutput.endsWith("/") || !tmpOutput.startsWith("\\")) tmpOutput += "\\";
            return tmpOutput;
        } else {
            return "/";
        }
    }

}
