package com.gk_software.core.maven.mojo;

import com.gk_software.core.maven.utils.SrcFiles;
import org.apache.logging.log4j.Logger;
import org.apache.maven.plugin.MojoExecutionException;
import org.apache.maven.plugin.MojoFailureException;
import org.apache.maven.plugins.annotations.LifecyclePhase;
import org.apache.maven.plugins.annotations.Mojo;
import org.apache.maven.plugins.annotations.Parameter;
import com.gk_software.core.maven.parser.JavaFile;
import com.gk_software.core.maven.utils.ClassData;
import com.gk_software.core.maven.utils.CustomConfigurationFactory;
import com.gk_software.core.maven.utils.ModificationHelper;
import org.codehaus.plexus.util.FileUtils;

import java.io.File;
import java.io.IOException;
import java.util.*;

@Mojo(name = "check-classes", defaultPhase = LifecyclePhase.INSTALL)
public class CheckClasses extends AbstractExecution {

    @Parameter(property = "classes")
    private List<String> classes;

    @Parameter(required = false, defaultValue = "")
    private String excludes;

    private Hashtable<String, ClassData> hashtable;

    private static Logger logger;

    private int possibleError = 0;

    private String[] regExExcludes = null;

    @Override
    public void execute() throws MojoExecutionException, MojoFailureException {
        long startTime = System.nanoTime();
        logger = CustomConfigurationFactory.getLogger(this.getOutput() + "check-classes");
        if ((!this.checkProductData()) || (!this.checkProjectData())) {
            return;
        }
        hashtable = new Hashtable<>();
        if (this.excludes != null) this.regExExcludes = ModificationHelper.excludesIntoPath(excludes);

        SrcFiles javaSrcPrjFiles = getProjectSrcFiles();
        javaSrcPrjFiles.setFiles(ModificationHelper.filterFiles(javaSrcPrjFiles.getFiles(), javaSuffix));


        LinkedList<File> owFiles = this.getChangedProductFiles();
        ModificationHelper.filterFiles(owFiles, javaSuffix);

        if (owFiles.size() == 0) {
            logger.info("There are no changed java files between product versions");
            getLog().info("There are no overwritten java files between product versions");
            return;
        }

        Hashtable<String, LinkedList<String>> productReferences = new Hashtable<>();
        LinkedList<File> filesToCompare = getClazzesFromConfig();
        LinkedList<JavaFile> javaFiles = getJavaFilesToCompare(filesToCompare, javaSrcPrjFiles.getFiles(), productReferences);
        getLog().info("Number of files to compare: " + javaFiles.size());

        for (JavaFile javaFile : javaFiles) {
            checkReferences(javaFile, owFiles, productReferences);
        }

        getLog().info("Number of java files that can be influenced by change in product: " + possibleError);
        addReferencesToHashTable(productReferences);
        printHashTable();

        long endTime = (System.nanoTime() - startTime) / 1000000;
        getLog().info("Execution time: " + endTime + "ms");
    }

    private void addReferencesToHashTable(Hashtable<String, LinkedList<String>> productReferences) {
        for (String key : hashtable.keySet()) {
            for (int index = 0; index < hashtable.get(key).getClasses().size(); index++) {
                checkWithProductReference(productReferences, key, index);
            }
        }
    }

    private void checkWithProductReference(Hashtable<String, LinkedList<String>> productReferences, String key, int index) {
        String productPath = hashtable.get(key).getClasses().get(index);
        for (String prdKey : productReferences.keySet()) {
            for (String reference : productReferences.get(prdKey)) {
                if (reference.equals(productPath)) {
                    hashtable.get(key).getClasses().set(index, (productPath + " defined in " + prdKey));
                    productReferences.get(prdKey).remove(reference);
                    return;
                }
            }
        }
    }

    private void printHashTable() {
        for (String key : hashtable.keySet()) {
            logger.info(ModificationHelper.getLineFromHashTable(key, hashtable.get(key)));
        }
    }

    private void checkReferences(JavaFile javaFile, LinkedList<File> owFiles, Hashtable<String, LinkedList<String>> productReferences) {
        Stack<String> stack = new Stack<>();
        checkOriginalFile(javaFile, (Hashtable<String, LinkedList<String>>) productReferences.clone());
        if (!javaFile.getSuperClass().equals("")) {
            stack.push(javaFile.getSuperClass());
        }

        String javaFRelativePath = ModificationHelper.getPathFromString(javaFile.getFile().getAbsolutePath(), this.getOldVersionPath());


        while (!stack.empty()) {
            String path = stack.pop();
            File oldProductFile = new File(this.getOldVersionPath() + path);

            if (!oldProductFile.exists()) {
                if (!ModificationHelper.isAmongExcludes(regExExcludes, path)) {
                    ModificationHelper.addToHashTable(hashtable, path, javaFRelativePath, "file not found");
                    possibleError++;
                }
            } else {
                if (checkIfChanged(hashtable, javaFRelativePath, path, owFiles, regExExcludes)) {
                    possibleError++;
                }
                String fileSuperClass = getJavaFileReference(oldProductFile);
                if (!fileSuperClass.equals("")) {
                    stack.push(fileSuperClass);
                }
            }
        }

    }

    private void checkOriginalFile(JavaFile javaFile, Hashtable<String, LinkedList<String>> productReferences) {
        String relativePath = ModificationHelper.getPathFromString(javaFile.getFile().getAbsolutePath(), this.getOldVersionPath());
        File oldFile = javaFile.getFile();
        File newFile = new File(this.getNewVersionPath() + relativePath);
        if (!newFile.exists()) {
            checkReference(productReferences, relativePath, "does not exists in new version");
            possibleError++;
        } else {
            try {
                if (!FileUtils.contentEquals(oldFile, newFile)) {
                    checkReference(productReferences, relativePath, "is changed in new version");
                    possibleError++;
                }
            } catch (IOException e) {
                getLog().warn(e.getMessage());
            }
        }

    }

    private void checkReference(Hashtable<String, LinkedList<String>> productReferences, String relativePath, String message) {
        for (String productRefKey : productReferences.keySet()) {
            for (String fileName : productReferences.get(productRefKey)) {
                if (fileName.compareTo(relativePath) == 0) {
                    logger.info(relativePath + " - " + message + " \n\t\t\t (defined in " + productRefKey + ")");
                    productReferences.get(productRefKey).remove(fileName);
                    return;
                }
            }
        }
        logger.info(relativePath + " - " + message);
    }


    private String getJavaFileReference(File file) {
        try {
            JavaFile jFile = new JavaFile(file);
            return jFile.getSuperClass();
        } catch (IOException e) {
            e.printStackTrace();
        }
        return "";
    }

    private LinkedList<JavaFile> getJavaFilesToCompare(LinkedList<File> filesToCompare, LinkedList<File> javaSrcPrjFiles, Hashtable<String, LinkedList<String>> productReferences) {
        LinkedList<JavaFile> javaFiles = new LinkedList<>();
        for (File f : filesToCompare) {
            try {
                JavaFile jFile = new JavaFile(f);
                javaFiles.add(jFile);
            } catch (IOException e) {
                System.out.println(e.getMessage());
            }
        }
        for (File f : javaSrcPrjFiles) {
            try {
                JavaFile jFile = new JavaFile(f);
                String relativePrjPath = ModificationHelper.getPathFromString(f.getAbsolutePath(), this.project.getBuild().getSourceDirectory());
                if (jFile.getReferences().size() > 0) {
                    productReferences.put(relativePrjPath, new LinkedList<String>());
                }
                for (String reference : jFile.getReferences()) {
                    File tmpFile = new File(this.getOldVersionPath() + reference);
                    if (tmpFile.exists()) {
                        productReferences.get(relativePrjPath).add(reference);
                        JavaFile tmpJFile = new JavaFile(tmpFile);
                        javaFiles.add(tmpJFile);
                    }
                }
            } catch (IOException e) {
                System.out.println(e.getMessage());
            }
        }
        return javaFiles;
    }

    private LinkedList<File> getClazzesFromConfig() {
        LinkedList<File> filesToCompare = new LinkedList<>();
        if (this.classes != null) {
            for (String s : this.classes) {
                String srcPath = ModificationHelper.getPathFromPackage(s, this.getOldVersionPath(), javaSuffix);
                File file = new File(srcPath);
                if (file.exists()) {
                    filesToCompare.add(file);
                } else {
                    getLog().warn(s + " was not found");
                }
            }
        }
        return filesToCompare;
    }

}


