package com.gk_software.core.maven.mojo;

import com.gk_software.core.maven.utils.SrcFiles;
import org.apache.logging.log4j.Logger;
import org.apache.maven.plugin.MojoExecutionException;
import org.apache.maven.plugin.MojoFailureException;
import org.apache.maven.plugins.annotations.LifecyclePhase;
import org.apache.maven.plugins.annotations.Mojo;
import org.apache.maven.plugins.annotations.Parameter;
import com.gk_software.core.maven.parser.PromoFile;
import com.gk_software.core.maven.utils.ClassData;
import com.gk_software.core.maven.utils.CustomConfigurationFactory;
import com.gk_software.core.maven.utils.ModificationHelper;

import java.io.File;
import java.io.IOException;
import java.util.Hashtable;
import java.util.LinkedList;
import java.util.Stack;

@Mojo(name = "check-promo", defaultPhase = LifecyclePhase.INSTALL)
public class CheckPromo extends AbstractExecution {

    @Parameter(required = false, defaultValue = "")
    private String excludes;

    private static final String suffix = ".promo";

    private static Logger logger;

    private String[] regExExcludes = null;

    private int possibleError = 0;

    private Hashtable<String, ClassData> hashtable = new Hashtable<>();

    @Override
    public void execute() throws MojoExecutionException, MojoFailureException {
        long startTime = System.nanoTime();
        logger = CustomConfigurationFactory.getLogger(this.getOutput() + "check-promo");
        if ((!this.checkProductData()) || (!this.checkProjectData())) {
            return;
        }
        if (this.excludes != null) this.regExExcludes = ModificationHelper.excludesIntoPath(excludes);

        SrcFiles projectPromoFiles = this.getProjectResourceFiles();
        projectPromoFiles.setFiles(ModificationHelper.filterFiles(projectPromoFiles.getFiles(), suffix));
        getLog().info("Promo files count: " + projectPromoFiles.getFiles().size());
        LinkedList<PromoFile> promoFiles = getPromoWithSuperClass(projectPromoFiles.getFiles());
        if (promoFiles.size() == 0) {
            logger.info("There were not found any promo files with superclass");
            getLog().info("There were not found any promo files with superclass");
            return;
        }
        getLog().info("Number of project promo files with superclass: "+promoFiles.size());
//        LinkedList<File> changedPromoFiles = this.getChangedProductFiles();
        LinkedList<File> changedPromoFiles = ModificationHelper.filterFiles(this.getChangedProductFiles(), suffix);
        getLog().info("Number of changed product promo files "+changedPromoFiles.size());
        for (PromoFile pFile : promoFiles) {
            checkPromoReference(pFile, changedPromoFiles);
        }
        getLog().info("Number of promo files that can be influenced by some change in product: "+ possibleError);
        printHashTable();

        long endTime = (System.nanoTime() - startTime) / 1000000;
        getLog().info("Execution time: " + endTime+"ms");
    }

    private void printHashTable() {
        for (String key: hashtable.keySet()){
            logger.info(ModificationHelper.getLineFromHashTable(key, hashtable.get(key)));
        }
    }

    private void checkPromoReference(PromoFile pFile, LinkedList<File> changedPromoFiles) {
        Stack<String> stack = new Stack<>();
        stack.push(pFile.getSuperClass());

        String promoRelativePath = ModificationHelper.getPathFromString(pFile.getFile().getAbsolutePath(), this.getProjectResourcePath());
        while (!stack.empty()) {
            String path = stack.pop();
            File oldFile = new File(this.getOldVersionPath() + path);
            if (!oldFile.exists()) {
                File projectFile = new File(this.getProjectResourcePath() + path);
                if (!projectFile.exists()){
                    if (!ModificationHelper.isAmongExcludes(regExExcludes, path)){
                        ModificationHelper.addToHashTable(hashtable, path, promoRelativePath, "file not found");
                        possibleError++;
                    }
                }else{
                    try {
                        PromoFile tmpPromoFile = new PromoFile(projectFile);
                        if (tmpPromoFile.getSuperClass() != null) {
                            stack.add(tmpPromoFile.getSuperClass());
                        }
                    } catch (IOException e) {
                        System.out.println(e.getMessage());
                    }
                }

            } else {
                if(checkIfChanged(hashtable, promoRelativePath, path, changedPromoFiles,regExExcludes)){
                    possibleError++;
                }

                try {
                    PromoFile tmpPromoFile = new PromoFile(oldFile);
                    if (tmpPromoFile.getSuperClass() != null) {
                        stack.add(tmpPromoFile.getSuperClass());
                    }
                } catch (IOException e) {
                    System.out.println(e.getMessage());
                }
            }

        }

    }




    private LinkedList<PromoFile> getPromoWithSuperClass(LinkedList<File> projectPromoFiles) {
        LinkedList<PromoFile> promoFiles = new LinkedList<>();
        for (File f : projectPromoFiles) {
            try {
                PromoFile tmpPromoFile = new PromoFile(f);
                if (tmpPromoFile.getSuperClass() != null) {
                    promoFiles.add(tmpPromoFile);
                }
            } catch (IOException e) {
                System.out.println(e.getMessage());
            }
        }
        return promoFiles;
    }
}
