package com.gk_software.core.maven.mojo;

import com.gk_software.core.maven.utils.SrcFiles;
import org.apache.logging.log4j.Logger;
import org.apache.maven.plugin.MojoExecutionException;
import org.apache.maven.plugin.MojoFailureException;
import org.apache.maven.plugins.annotations.LifecyclePhase;
import org.apache.maven.plugins.annotations.Mojo;
import org.apache.maven.plugins.annotations.Parameter;
import com.gk_software.core.maven.parser.PromoJFile;
import com.gk_software.core.maven.utils.ClassData;
import com.gk_software.core.maven.utils.CustomConfigurationFactory;
import com.gk_software.core.maven.utils.ModificationHelper;

import java.io.File;
import java.io.IOException;
import java.util.Hashtable;
import java.util.LinkedList;
import java.util.Stack;

@Mojo(name = "check-promoj", defaultPhase = LifecyclePhase.INSTALL)
public class CheckPromoj extends AbstractExecution {

    @Parameter(required = false, defaultValue = "")
    private String excludes;

    private static final String suffix = ".promoj";

    private static Logger logger;

    private String[] regExExcludes = null;

    private int possibleError = 0;

    private Hashtable<String, ClassData> hashtable = new Hashtable<>();

    @Override
    public void execute() throws MojoExecutionException, MojoFailureException {
        long startTime = System.nanoTime();
        logger = CustomConfigurationFactory.getLogger(this.getOutput() + "check-promoj");
        if ((!this.checkProductData()) || (!this.checkProjectData())) {
            return;
        }
        if (this.excludes != null) this.regExExcludes = ModificationHelper.excludesIntoPath(excludes);

        SrcFiles projectPromojFiles = this.getProjectResourceFiles();
        projectPromojFiles.setFiles(ModificationHelper.filterFiles(projectPromojFiles.getFiles(), suffix));
        getLog().info("Promoj files count: " + projectPromojFiles.getFiles().size());

        LinkedList<PromoJFile> promojFiles = getPromoJWithSuperClass(projectPromojFiles.getFiles());
        if (promojFiles.size() == 0) {
            logger.info("There were not found any promoj files with superclass");
            getLog().info("There were not found any promoj files with superclass");
            return;
        }
        getLog().info("Number of project promoj files with superclass: " + promojFiles.size());
        LinkedList<File> changedPromojFiles = ModificationHelper.filterFiles(this.getChangedProductFiles(), suffix);

        for (PromoJFile pFile : promojFiles) {
            checkPromoReference(pFile, changedPromojFiles);
        }
        getLog().info("Number of promoj files that can be influenced by some change in product: " + possibleError);
        printHashTable();

        long endTime = (System.nanoTime() - startTime) / 1000000;
        getLog().info("Execution time: " + endTime+"ms");
    }

    private LinkedList<PromoJFile> getPromoJWithSuperClass(LinkedList<File> projectPromoFiles) {
        LinkedList<PromoJFile> promojFiles = new LinkedList<>();
        for (File f : projectPromoFiles) {
            try {
                PromoJFile tmpPromoFile = new PromoJFile(f);
                if (tmpPromoFile.getSuperClass() != null) {
                    promojFiles.add(tmpPromoFile);
                }
            } catch (IOException e) {
                System.out.println(e.getMessage());
            }
        }
        return promojFiles;
    }


    private void printHashTable() {
        for (String key : hashtable.keySet()) {
            logger.info(ModificationHelper.getLineFromHashTable(key, hashtable.get(key)));
        }
    }

    private void checkPromoReference(PromoJFile pFile, LinkedList<File> changedPromoFiles) {
        Stack<String> stack = new Stack<>();
        stack.push(pFile.getSuperClass());

        String promoRelativePath = ModificationHelper.getPathFromString(pFile.getFile().getAbsolutePath(), this.getProjectResourcePath());
        while (!stack.empty()) {
            String path = stack.pop();
            File oldFile = new File(this.getOldVersionPath() + path);
            if (!oldFile.exists()) {
                File projectFile = new File(this.getProjectResourcePath()+path);
                if (!projectFile.exists()){
                    if (!ModificationHelper.isAmongExcludes(regExExcludes,path)){
                        ModificationHelper.addToHashTable(hashtable, path, promoRelativePath, "file not found");
                        possibleError++;
                    }
                }else{
                    try {
                        PromoJFile tmpPromoFile = new PromoJFile(projectFile);
                        if (tmpPromoFile.getSuperClass() != null) {
                            stack.add(tmpPromoFile.getSuperClass());
                        }
                    } catch (IOException e) {
                        System.out.println(e.getMessage());
                    }
                }

            } else {
                if (checkIfChanged(hashtable, promoRelativePath, path, changedPromoFiles, regExExcludes)) {
                    possibleError++;
                }
                try {
                    PromoJFile tmpPromoFile = new PromoJFile(oldFile);
                    if (tmpPromoFile.getSuperClass() != null) {
                        stack.add(tmpPromoFile.getSuperClass());
                    }
                } catch (IOException e) {
                    System.out.println(e.getMessage());
                }
            }

        }

    }
}
