package com.gk_software.core.maven.mojo;

import org.apache.logging.log4j.Logger;
import org.apache.maven.plugin.MojoExecutionException;
import org.apache.maven.plugin.MojoFailureException;
import org.apache.maven.plugins.annotations.LifecyclePhase;
import org.apache.maven.plugins.annotations.Mojo;
import org.apache.maven.plugins.annotations.Parameter;
import org.codehaus.plexus.util.FileUtils;
import com.gk_software.core.maven.utils.CustomConfigurationFactory;
import com.gk_software.core.maven.utils.ModificationHelper;


import java.io.File;
import java.io.IOException;
import java.util.Hashtable;
import java.util.LinkedList;

@Mojo(name = "overwritten-files", defaultPhase = LifecyclePhase.INSTALL)
public class OverwrittenFiles extends AbstractExecution {

    @Parameter(required = false, defaultValue = "")
    private String includes;

    @Parameter(required = false, defaultValue = "")
    private String excludes;

    @Parameter
    private boolean showAllOverwritten;


    private String[] regExIncludes = null;
    private String[] regExExcludes = null;

    private static Logger logger;

    @Override
    public void execute() throws MojoExecutionException, MojoFailureException {
        long startTime = System.nanoTime();

        logger = CustomConfigurationFactory.getLogger(this.getOutput() + "overwritten-files");
        if ((!this.checkProductData()) || (!this.checkProjectData())) {
            return;
        }

        if (this.includes != null) this.regExIncludes = ModificationHelper.strToRegex(includes);
        if (this.excludes != null) this.regExExcludes = ModificationHelper.excludesIntoPath(excludes);

        LinkedList<File> oldProductFiles = this.getOldVersionFiles();

        getLog().info("Number of old product files: " + oldProductFiles.size());
        getLog().info("Number of new product files: " + this.getNewVersionFiles().size());

        Hashtable<String, LinkedList<File>> projectFiles = this.getAllSrcFiles();

        if (regExIncludes != null) {
            for (String key : projectFiles.keySet()) {
                projectFiles.put(key, ModificationHelper.filterFiles(projectFiles.get(key), regExIncludes));
            }
        }

        Hashtable<String, LinkedList<File>> owProjectFiles = findOwProjectFiles(projectFiles, oldProductFiles);

        LinkedList<File> changedProductFiles = this.getChangedProductFiles();

        getLog().info("Number of changed files between product versions: " + changedProductFiles.size());

        int owPrjFilesCount = solveOverwrittenFiles(owProjectFiles, changedProductFiles);

        if (owPrjFilesCount == 0) {
            getLog().info("There are no overwritten project files");
            logger.info("There are no overwritten project files");
        } else {
            getLog().info("There are " + owPrjFilesCount + " overwritten project files");
        }

        long endTime = (System.nanoTime() - startTime) / 1000000;
        getLog().info("Execution time: " + endTime + "ms");
    }

    private int solveOverwrittenFiles(Hashtable<String, LinkedList<File>> owProjectFiles, LinkedList<File> changedProductFiles) {
        int owFilesCount = 0;
        for (String key : owProjectFiles.keySet()) {
            for (File owFile : owProjectFiles.get(key)) {
                String relativePath = ModificationHelper.getPathFromString(owFile.getAbsolutePath(), key);
                File newVFile = new File(this.getNewVersionPath() + relativePath);
                if (!newVFile.exists()) {
                    if (!ModificationHelper.isAmongExcludes(regExExcludes, relativePath)) {
                        logger.info(relativePath + " - is overwritten and does not exists in new version");
                        owFilesCount++;
                    }
                    continue;
                }
                boolean changedBool = false;
                for (File changed : changedProductFiles) {
                    if (changed.getAbsolutePath().endsWith(relativePath)) {
                        if (!ModificationHelper.isAmongExcludes(regExExcludes, relativePath)) {
                            logger.info(relativePath + " - is overwritten in changed product files");
                            owFilesCount++;
                        }
                        changedBool = true;
                        break;
                    }
                }
                if (showAllOverwritten){
                    if (!changedBool) {
                        if (!ModificationHelper.isAmongExcludes(regExExcludes, relativePath)) {
                            owFilesCount++;
                            logger.info(relativePath + " - is overwritten");
                        }
                    }
                }
            }
        }
        return owFilesCount;
    }

    private Hashtable<String, LinkedList<File>> findOwProjectFiles(Hashtable<String, LinkedList<File>> projectFiles, LinkedList<File> oldProductFiles) {
        Hashtable<String, LinkedList<File>> hashtable = new Hashtable<>();
        for (String key : projectFiles.keySet()) {
            for (File prjSrcF : projectFiles.get(key)) {
                String relativePrjPath = ModificationHelper.getPathFromString(prjSrcF.getAbsolutePath(), key);
                for (File oldF : oldProductFiles) {
                    if (oldF.getAbsolutePath().endsWith(relativePrjPath)) {
                        try {
                            if (!FileUtils.contentEquals(oldF, prjSrcF)) {
                                if (!hashtable.containsKey(key)) {
                                    hashtable.put(key, new LinkedList<File>());
                                }
                                hashtable.get(key).add(prjSrcF);
                            }
                        } catch (IOException e) {
                            System.out.println(e.getMessage());
                        }
                    }
                }
            }
        }
        return hashtable;
    }
}