package com.gk_software.core.maven.parser;

import com.gk_software.core.maven.utils.ModificationHelper;

import java.io.*;
import java.util.LinkedList;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

public abstract class AbstractParser {

    protected static final int END_READING = 0;
    protected static final int CONTINUE_READING = 1;

    protected File file;
    protected LinkedList<String> imports;
    protected String superClass;

    private static final Pattern IMPORT_PATTERN = Pattern.compile("(?:Import)\\s([a-zA-Z0-9_\\-.]+)");
    private static final Pattern PACKAGE_PATTERN = Pattern.compile("(?:Package )([a-zA-Z0-9_\\-.,\"]+)");

    private boolean commented;

    public AbstractParser(File file) throws IOException {
        BufferedReader reader;
        reader = new BufferedReader(new FileReader(file.getAbsolutePath()));
        this.file = file;
        imports = new LinkedList<>();
        String line = reader.readLine();
        while (line != null) {
            line = line.trim();
            if (parseData(line) == END_READING) {
                break;
            }
            line = reader.readLine();
        }
        reader.close();
    }

    protected int parseData(String line) {
        if (lineIsComment(line.trim())) return CONTINUE_READING;
        Matcher m = getImportPattern().matcher(line);
        if (m.find()) {
            String tmpImport = m.group(1);
            imports.add(tmpImport);
            return CONTINUE_READING;
        }
        m = getClassPattern().matcher(line);
        if (m.find()) {
            m = extendsPattern().matcher(line);
            if (m.find()) {
                String extendedClass = getSuperClassFromLine(m.group(0));
                superClass = addSuperClassPath(extendedClass);
            }
            return END_READING;
        }
        return CONTINUE_READING;
    }


    private boolean lineIsComment(String line) {
        if (commented) {
            if (line.endsWith("*/")) commented = false;
            return true;
        }
        if (line.startsWith("/*")) {
            if (!line.endsWith("*/")) commented = true;
            return true;
        }

        return line.startsWith("//");
    }


    protected String addSuperClassPath(String superClass) {
        StringBuilder sb = new StringBuilder();
        sb.append("\\");
        if (superClass.contains(".")) {
            sb.append(superClass.replace(".", "\\"));
            sb.append(getSuffix());
            return sb.toString();
        } else {
            File inPackage = new File(file.getAbsolutePath().replace(file.getName(), superClass) + getSuffix());
            if (inPackage.exists()) {
                sb = new StringBuilder();
                String path = getPackage();
                if (!path.equals("")) {
                    sb.append(path).append("\\");
                } else {
                    sb.append("\\");
                }
                sb.append(superClass).append(getSuffix());
                return sb.toString();
            }

            for (String imp : imports) {
                String[] tokens = imp.split("\\.");
                if (tokens[tokens.length - 1].equals(superClass)) {
                    sb.append(imp.replace(".", "\\"));
                    sb.append(getSuffix());
                    return sb.toString();
                }
            }
        }

        return sb.toString();
    }


    private String getPackage() {
        String packagePath = "";
        try {
            BufferedReader reader = new BufferedReader(new FileReader(file.getAbsolutePath()));
            String line = reader.readLine();
            Matcher m;
            while (line != null) {
                line = line.trim();
                m = getPackagePattern().matcher(line);
                if (m.find()) {
                    packagePath = ModificationHelper.getPathFromPackage(m.group(1));
                break;
            }
            line = reader.readLine();
            }
            reader.close();
        } catch (IOException e) {
            System.out.println(e.getMessage());
        }
        return packagePath;
    }

    protected abstract String getSuperClassFromLine(String line);

    protected Pattern getImportPattern() {
        return IMPORT_PATTERN;
    }

    protected Pattern getPackagePattern() {
        return PACKAGE_PATTERN;
    }

    protected abstract String getSuffix();

    protected abstract Pattern extendsPattern();

    protected abstract Pattern getClassPattern();

    public File getFile() {
        return file;
    }

    public LinkedList<String> getImports() {
        return imports;
    }

    public String getSuperClass() {
        return superClass;
    }
}
