package com.gk_software.core.maven.parser;

import com.gk_software.core.maven.utils.ModificationHelper;

import java.io.*;
import java.util.LinkedList;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

public class JavaFile {
    private static final int END_READING = 0;
    private static final int CONTINUE_READING = 1;

    private File file;
    private List<String> references;
    private List<String> imports;


    private boolean productReference;
    private boolean commented;
    private String superClass;

    private static final Pattern IMPORT_PATTERN = Pattern.compile("(?:import)\\s([a-zA-Z0-9_\\-.]+)");
    private static final Pattern CLASS_PATTERN = Pattern.compile("((?:public class)|(?:public interface)|(?:public abstract))\\s([a-zA-Z0-9_\\-.]+)");
    private static final Pattern EXTENDS_PATTERN = Pattern.compile("(?:public class)\\s([a-zA-Z0-9_\\-.]+)\\s(?:extends)\\s([a-zA-Z0-9_\\-.]+)");
    private static final Pattern ANNOTATION_PATTERN = Pattern.compile("(?:@ProductReference)([a-zA-Z0-9_\\-.(), \"]+)");
    private static final Pattern PACKAGE_PATTERN = Pattern.compile("(?:package )([a-zA-Z0-9_\\-.,\"]+)");

    public JavaFile(File file) throws IOException {
        BufferedReader reader;
        reader = new BufferedReader(new FileReader(file.getAbsolutePath()));
        this.imports = new LinkedList<>();
        this.references = new LinkedList<>();
        this.superClass = "";
        this.file = file;
        this.productReference = false;

        String line = reader.readLine();
        while (line != null) {
            line = line.trim();
            if (parseData(line) == END_READING) {
                break;
            }
            line = reader.readLine();
        }
        reader.close();
    }


    private String addSuperClassPath(String superClass) {
        StringBuilder sb = new StringBuilder();
        sb.append("\\");
        if (superClass.contains(".")) {
            sb.append(superClass.replace(".", "\\"));
            sb.append(".java");
            return sb.toString();
        } else {
            File inPackage = new File(file.getAbsolutePath().replace(file.getName(), superClass) + ".java");
            if (inPackage.exists()) {
                sb = new StringBuilder();
                String path = getPackage();
                if (!path.equals("")) {
                    sb.append(path).append("\\");
                } else {
                    sb.append("\\");
                }
                sb.append(superClass).append(".java");
                return sb.toString();
            }
            for (String imp : imports) {
                String[] tokens = imp.split("\\.");
                if (tokens[tokens.length - 1].equals(superClass)) {
                    sb.append(imp.replace(".", "\\"));
                    sb.append(".java");
                    return sb.toString();
                }
            }
        }

        return sb.toString();
    }

    private int parseData(String line) {
        if (lineIsComment(line)) return CONTINUE_READING;
        Matcher m = IMPORT_PATTERN.matcher(line);
        if (m.find()) {
            String tmpImport = m.group(0).replace("import ", "");
            imports.add(tmpImport);
        }
        m = ANNOTATION_PATTERN.matcher(line);
        if (m.find()) {
            this.productReference = true;
            String reference;
            if (m.group(0).endsWith("\")")) {
                reference = m.group(0).substring(19, m.group().length() - 2);
            } else {
                reference = m.group(0).substring(18, m.group().length() - 1);
            }
            if (reference != null) {
                String[] tokens = reference.split(",");
                for (String token : tokens) {
                    references.add(ModificationHelper.getPathFromPackage(token.trim(), ".java"));
                }
            }
        }
        m = CLASS_PATTERN.matcher(line);
        if (m.find()) {
            parseSuperClasses(line);
            return END_READING;
        }

        return CONTINUE_READING;
    }

    private boolean lineIsComment(String line) {
        if (commented) {
            if (line.endsWith("*/")) commented = false;
            return true;
        }
        if (line.startsWith("/*")) {
            if (!line.endsWith("*/")) commented = true;
            return true;
        }
        if (line.startsWith("//")) {
            Matcher m = ANNOTATION_PATTERN.matcher(line);
            return !m.find();
        }

        return false;
    }

    private void parseSuperClasses(String line) {
        Matcher m = EXTENDS_PATTERN.matcher(line);
        if (m.find()) {
            String[] tokens = m.group(0).split("extends");
            String extendedClass = tokens[1].trim();
            superClass = addSuperClassPath(extendedClass.replace("{", ""));
        }
    }

    private String getPackage() {
        String packagePath = "";
        try {
            BufferedReader reader = new BufferedReader(new FileReader(file.getAbsolutePath()));
            String line = reader.readLine();
            Matcher m;
            while (line != null) {
                line = line.trim();
                m = PACKAGE_PATTERN.matcher(line);
                if (m.find()) {
                    packagePath = ModificationHelper.getPathFromPackage(m.group(1));
                    break;
                }
                line = reader.readLine();
            }
            reader.close();
        } catch (IOException e) {
            System.out.println(e.getMessage());
        }
        return packagePath;
    }

    public List<String> getImports() {
        return this.imports;
    }

    public String getSuperClass() {
        return superClass;
    }

    public File getFile() {
        return file;
    }

    public List<String> getReferences() {
        return references;
    }

    public boolean isProductReference() {
        return productReference;
    }
}
